package net.grelf.grip;

// This Java source file is part of the application GRIP, for processing and
// measuring digital photographs.
// GRIP is available free from www.grelf.net - please always quote this URL in
// connection with GRIP.
//
// Copyright (c) Graham T Relf, UK, 2006
//
// All software contains bugs, even if, we hope, they are only extremely subtle
// ones. The author has tested this software to the best of his abilities to
// ensure that it is generally fit for its purpose of processing and measuring
// digital photographs. The author makes absolutely no claims and gives no
// guarantees whatsoever as to the accuracy of any output from this software.
// Anyone making decisions based on any output from this software has a
// responsibility to first convince themselves of the appropriateness of so
// doing. They should test the software for themselves and understand how it
// works, sufficiently for their needs. Source code and explanatory
// documentation are available for that purpose.

/** Batch menu. */
public class BatchMenu
extends javax.swing.JMenu
implements java.awt.event.ActionListener, javax.swing.event.MenuListener,
           ImageSizeAction
{
	private javax.swing.JMenuItem itemAccum, itemCombine, itemWarp, itemMerge,
                                itemComet, itemMeas, itemTrail, itemAnim,
                                itemMedian, itemHdr, itemMeteor;

	public BatchMenu ()
	{
		super ("Batch");
    setMnemonic ('B');
    initialise ();
  } // BatchMenu

  private void initialise ()
  {
    add (net.grelf.Util.menuItem (this,
            "Image information (as XML file)", "infoxml",
            'X', 0));
    add (net.grelf.Util.menuItem (this,
            "Slide show", "slides",
            'W', 0));
    addSeparator ();

    javax.swing.JMenu astroSubmenu = new javax.swing.JMenu ("Astro");
    astroSubmenu.add (itemCombine =
            net.grelf.Util.menuItem (this,
            "Combine into 1 image (stack)", "astrocombine",
            '1', 0));
		itemCombine.setToolTipText (
            "Warp or shift to same positions and accumulate (formerly Astro-process)");
    astroSubmenu.add (itemComet =
            net.grelf.Util.menuItem (this,
            "Combine (stack) for comets", "comet",
            'A', 0));
		itemComet.setToolTipText (
            "Semi-interactive: point to object in every image to keep it still");
    astroSubmenu.add (itemWarp =
            net.grelf.Util.menuItem (this,
            "Warp/shift onto common basis", "astrowarp",
            'O', 0));
		itemWarp.setToolTipText (
            "Blobs will be in same positions on every _warp or _shav image");
    astroSubmenu.add (itemTrail =
            net.grelf.Util.menuItem (this,
            "Accumulate star trails", "astrotrail",
            0, 0));
		itemTrail.setToolTipText ("Star brightnesses are boosted");
    astroSubmenu.add (itemAnim =
            net.grelf.Util.menuItem (this,
            "Animate at 100% scale", "astroanim",
            0, 0));
		itemAnim.setToolTipText ("You need to have warped/shifted onto common basis first");
    astroSubmenu.add (itemMedian =
            net.grelf.Util.menuItem (this,
            "Create median pixels image", "astromedian",
            0, 0));
		itemMedian.setToolTipText (
            "You need to have warped/shifted onto common basis first");
    astroSubmenu.add (itemMeas =
            net.grelf.Util.menuItem (this,
            "Measure star(s) (select on first image)", "astromeas",
            'E', 0));
		itemMeas.setToolTipText ("You need to have warped/shifted onto common basis first");
		astroSubmenu.add (itemMerge =
            net.grelf.Util.menuItem (this,
            "Merge star charts", "charts",
            'G', 0));
		astroSubmenu.add (itemMeteor =
            net.grelf.Util.menuItem (this,
            "Find meteors", "meteors",
            '0', 0));
    add (astroSubmenu);

    javax.swing.JMenu geomSubmenu = new javax.swing.JMenu ("Geometry");
    geomSubmenu.add (net.grelf.Util.menuItem (this,
            "Fit images in m x n", "fit",
            'F', 0));
    geomSubmenu.add (net.grelf.Util.menuItem (this,
            "Crop (set on first image)", "crop",
            'P', 0));
    geomSubmenu.add (net.grelf.Util.menuItem (this,
            "Flip horizontally", "fliph",
            'Z', 0));
    geomSubmenu.add (net.grelf.Util.menuItem (this,
            "Flip vertically", "flipv",
            'Y', 0));
    geomSubmenu.add (net.grelf.Util.menuItem (this,
            "Scale (interpolated)", "scale",
            'L', 0));
    geomSubmenu.add (net.grelf.Util.menuItem (this,
            "Magnify (replicated)", "magnify",
            0, 0));
    geomSubmenu.add (net.grelf.Util.menuItem (this,
            "Rotate", "rotate",
            'R', 0));
    geomSubmenu.add (net.grelf.Util.menuItem (this,
            "Inverse gnomonic projection", "invgnomonic",
            'J', 0));
    geomSubmenu.add (net.grelf.Util.menuItem (this,
            "Gnomonic projection", "gnomonic",
            'N', 0));
		geomSubmenu.add (net.grelf.Util.menuItem (this,
            "Apply warping grid", "warpgrid",
            0, 0));
    add (geomSubmenu);

    javax.swing.JMenu levelsSubmenu = new javax.swing.JMenu ("Levels");
    levelsSubmenu.add (net.grelf.Util.menuItem (this, "Invert contrast", "invert",
            'I', 0));
    levelsSubmenu.add (net.grelf.Util.menuItem (this,
            "Contrast auto-stretch", "autostretch",
            'T', 0));
    levelsSubmenu.add (net.grelf.Util.menuItem (this, "Convert to monochrome", "mono",
            'C', 0));
    levelsSubmenu.add (net.grelf.Util.menuItem (this,
            "Convert to 8 bits per channel", "8bit",
            '8', 0));
    levelsSubmenu.add (net.grelf.Util.menuItem (this,
            "Convert to 16 bits per channel", "16bit",
            '6', 0));
    levelsSubmenu.add (net.grelf.Util.menuItem (this,
            "Convert to 32 bits per channel", "32bit",
            '3', 0));
    levelsSubmenu.add (net.grelf.Util.menuItem (this,
            "Convert to 64 bits per channel", "64bit",
            '4', 0));
    levelsSubmenu.add (net.grelf.Util.menuItem (this,
            "Correct background", "background",
            'B', 0));
    levelsSubmenu.add (net.grelf.Util.menuItem (this, "Saturate", "saturate",
            0, 0));
    levelsSubmenu.add (net.grelf.Util.menuItem (this,
            "Square (multiply by self)", "square",
            'Q', 0));
    add (levelsSubmenu);

    javax.swing.JMenu combineSubmenu = new javax.swing.JMenu ("Combine with an image");
    combineSubmenu.add (net.grelf.Util.menuItem (this,
            "Subtract an image", "subtract",
            'U', 0));
    combineSubmenu.add (net.grelf.Util.menuItem (this,
            "Multiply by an image", "multiply",
            'M', 0));
    combineSubmenu.add (net.grelf.Util.menuItem (this,
            "Divide by a flat-field image", "divide",
            'D', 0));
    add (combineSubmenu);

    javax.swing.JMenu formatSubmenu = new javax.swing.JMenu ("Convert file format");
    formatSubmenu.add (net.grelf.Util.menuItem (this,
            "to 64-bit FITS", "FITS64", 0, 0));
    formatSubmenu.add (net.grelf.Util.menuItem (this,
            "to 32-bit FITS", "FITS32", 0, 0));
    formatSubmenu.add (net.grelf.Util.menuItem (this,
            "to 16-bit FITS", "FITS16", 0, 0));
    formatSubmenu.add (net.grelf.Util.menuItem (this,
            "to 8-bit FITS", "FITS8", 0, 0));
    formatSubmenu.add (net.grelf.Util.menuItem (this,
            "to 32-bit TIFF (for Photoshop only)", "TIFF32", 0, 0));
    formatSubmenu.add (net.grelf.Util.menuItem (this,
            "to 16-bit TIFF", "TIFF16", 0, 0));
    formatSubmenu.add (net.grelf.Util.menuItem (this,
            "to 8-bit TIFF", "TIFF8", 0, 0));
    formatSubmenu.add (net.grelf.Util.menuItem (this,
            "to JPEG (8bits, lossy)", "JPEG8", 0, 0));
    add (formatSubmenu);

    addSeparator ();
    add (net.grelf.Util.menuItem (this,
            "Average images into 1", "average",
            'V', 0));
    add (itemAccum = net.grelf.Util.menuItem (this,
            "Accumulate images into 1", "accumulate",
            0, 0));
		itemAccum.setToolTipText (
            "Like averaging but uses more bits and does not scale down");
    add (net.grelf.Util.menuItem (this,
            "Interpret RAW & save as TIFF", "raw",
            '&', 0));
/****
    add (net.grelf.Util.menuItem (this,
            "Make HDR image", "hdr",
            0, 0));
****/
    addSeparator ();
    add (net.grelf.Util.menuItem (this,
            "Make your own process", "diy",
            'K', 0));
    addSeparator ();
    add (net.grelf.Util.menuItem (this,
            "Help about this menu", "help",
            'H', 0));
	} // initialise

	/** Implementing ImageSizeAction to allow call-back from the "fit" process. */
	@Override
	public void process (ImageSizeDialogue isd)
	{
		java.util.Map <String, Object> params = new java.util.HashMap <> ();
		int width = isd.getWd ();
		int height = isd.getHt ();
		isd.dispose ();
		params.put ("width", width);
		params.put ("height", height);

		if (0 < width && 0 < height)
		{
			new BatchProcessor (BatchAction.FIT).
                          batch ("_" +	width + "x" + height, params);
		}
	} // process

	/** Implementing the ActionListener interface to act on menu option selection. */
	@Override
	public void actionPerformed (java.awt.event.ActionEvent ev)
	{
		String action = ev.getActionCommand ();
		java.util.Map <String, Object> params = new java.util.HashMap <> ();

    switch (action)
    {
      case "8bit":
        new BatchProcessor (BatchAction.BIT8).batch ("_8bit", params);
        break;

      case "16bit":
        new BatchProcessor (BatchAction.BIT16).batch ("_16bit", params);
        break;

      case "32bit":
        new BatchProcessor (BatchAction.BIT32).batch ("_32bit", params);
        break;

      case "64bit":
        new BatchProcessor (BatchAction.BIT64).batch ("_64bit", params);
        break;

      case "accumulate":
        new BatchProcessor (BatchAction.ACCUM).batch ("_acc", params);
        break;

      case "astroanim":
        new net.grelf.sequence.Animator ().process ();
        break;

      case "astrocombine":
        new AstroProcessDialogue (GRIP.getInstance (), AstroProcessKind.COMBINE);
        break;

      case "astromedian":
        if (net.grelf.Util.confirm ("Confirm", "Are the images aligned?"))
        {
          new net.grelf.sequence.Medianiser ().process ();
        }

        break;

      case "astromeas":
        if (net.grelf.Util.confirm ("Prerequisite",
                "Have you aligned the images\n(option Warp/shift onto common basis)?"))
        {
          new StarMeasProcessor ().run ();
        }
        else
        {
          new AstroProcessDialogue (GRIP.getInstance (), AstroProcessKind.ALIGN);
        }
        break;

      case "astrotrail":
        new net.grelf.sequence.Maximiser ().process ();
//        new BatchProcessor ().batch ("trail", "_trail", params);
        break;

      case "astrowarp":
        new AstroProcessDialogue (GRIP.getInstance (), AstroProcessKind.ALIGN);
        break;

      case "autostretch":
        boolean allSame = net.grelf.Util.confirm (null, "Contrast stretch",
                "All channels stretched the same?");
        params.put ("allSame", allSame);
        new BatchProcessor (BatchAction.AUTOSTRETCH).batch ("_stretch", params);
        break;

      case "average":
        new BatchProcessor (BatchAction.AVERAGE).batch ("_ave", params);
        break;

      case "background":
        new BatchProcessor (BatchAction.BACKGROUND).batch ("_bkgr", params);
        break;

      case "charts":
        java.util.List <net.grelf.grip.StarChartFrame> chartFrames =
                net.grelf.grip.GRIP.getInstance ().getAllStarChartFrames ();
        net.grelf.astro.StarChart.mergeCharts (chartFrames);
        break;

      case "comet":
        new AstroProcessDialogue (GRIP.getInstance (), AstroProcessKind.COMBINE,
                true); // Interactive
        break;

      case "crop":
        new BatchProcessor (BatchAction.CROP).batch ("_crop", params);
        break;

      case "divide":
        String divPath = net.grelf.FileIO.selectInputFile ("Flat field", null,
                                new javax.swing.filechooser.FileFilter [] {
                                new net.grelf.FileIOImageFilter () });
        if (null != divPath)
        {
          params.put ("divPath", divPath);
          new BatchProcessor (BatchAction.DIVIDE).batch ("_div", params);
        }

        break;

      case "diy":
        new BatchDialogue ();
        break;

      case "fit":
        int width = 1024;
        int height = 768;
        new ImageSizeDialogue (width, height, "Scale to fit", this);
        // Will invoke BatchProcessor
        break;

      case "FITS64":
        new net.grelf.sequence.Reformatter ().
            reformat (net.grelf.image.FileFormat.FITS, 64);
        break;

      case "FITS32":
        new net.grelf.sequence.Reformatter ().
            reformat (net.grelf.image.FileFormat.FITS, 32);
        break;

      case "FITS16":
        new net.grelf.sequence.Reformatter ().
            reformat (net.grelf.image.FileFormat.FITS, 16);
        break;

      case "FITS8":
        new net.grelf.sequence.Reformatter ().
            reformat (net.grelf.image.FileFormat.FITS, 8);
        break;

      case "fliph":
        new BatchProcessor (BatchAction.FLIPH).batch ("_flh", params);
        break;

      case "flipv":
        new BatchProcessor (BatchAction.FLIPV).batch ("_flv", params);
        break;

      case "gnomonic":
        new BatchProcessor (BatchAction.GNOMONIC).batch ("_gnm", params);
        break;

      case "hdr":
        net.grelf.hdr.HDR.buildHdrImage ();
        break;

      case "help":
        HelpMenu.help ("pages/batch_menu.html");
        break;

      case "infoxml":
        String outputPath = net.grelf.FileIO.selectOutputFile ("Save as XML file", null,
              new net.grelf.FileIO_XMLFilter [] {
              new net.grelf.FileIO_XMLFilter (net.grelf.FileIO_XMLFilter.XML) },
              "x.xml");

        if (null != outputPath)
        {
          params.put ("outputPath", outputPath);
          new BatchProcessor (BatchAction.INFOXML).batch (null, params);
        }

        break;

      case "invert":
        new BatchProcessor (BatchAction.INVERT).batch ("_inv", params);
        break;

      case "invgnomonic":
        new BatchProcessor (BatchAction.INVGNOMONIC).batch ("_ignm", params);
        break;

      case "JPEG8":
        new net.grelf.sequence.Reformatter ().
            reformat (net.grelf.image.FileFormat.JPEG, 8);
        break;

      case "magnify":
        {
          int factor = net.grelf.Util.askInteger (
                  "Factor to magnify by [2..8]?", 2, 2, 8);

          if (2 <= factor && factor <= 8) // User didn't cancel
          {
            params.put ("factor", factor);
            new BatchProcessor (BatchAction.MAGNIFY).batch ("_mgn", params);
          }

          break;
        }
      
      case "meteors":
        new net.grelf.meteor.MeteorDetector ().run ();
        break;

      case "mono":
        new BatchProcessor (BatchAction.MONO).batch ("_mono", params);
        break;

      case "multiply":
        String mulPath = 
            net.grelf.FileIO.selectInputFile ("Image to multiply by", null,
                                new javax.swing.filechooser.FileFilter [] {
                                        new net.grelf.FileIOImageFilter () });
        if (null != mulPath)
        {
          params.put ("mulPath", mulPath);
          new BatchProcessor (BatchAction.MULTIPLY).batch ("_mpy", params);
        }

        break;

      case "raw":
        new BatchProcessor (BatchAction.RAW).batch ("_", params);
        break;

      case "rotate":
        double angle = net.grelf.Util.askDouble (
                "Angle to rotate anti-clockwise (degrees) [-360..360]?",
                0.0, -360.0, 360.0);

        if (-360.0 <= angle && angle <= 360.0) // User didn't cancel
        {
          params.put ("angle", angle);
          new BatchProcessor (BatchAction.ROTATE).batch ("_rot", params);
        }

        break;

      case "saturate":
        new BatchProcessor (BatchAction.SATURATE).batch ("_sat", params);
        break;

      case "scale":
        {
          double factor = net.grelf.Util.askDouble (
                  "Factor to scale by [0.1..10]?", 1.0, 0.1, 10.0);

          if (0.1 <= factor && factor <= 10.0) // User didn't cancel
          {
            params.put ("factor", factor);
            new BatchProcessor (BatchAction.SCALE).batch ("_scl", params);
          }

          break;
        }

      case "slides":
        slideShow ();
        break;

      case "square":
        new BatchProcessor (BatchAction.SQUARE).batch ("_squ", params);
        break;

      case "subtract":
        String subPath = 
            net.grelf.FileIO.selectInputFile ("Image to subtract", null,
                                   new javax.swing.filechooser.FileFilter [] {
                                   new net.grelf.FileIOImageFilter () });
        if (null != subPath)
        {
          params.put ("subPath", subPath);
          new BatchProcessor (BatchAction.SUBTRACT).batch ("_sub", params);
        }

        break;

      case "TIFF32":
        new net.grelf.sequence.Reformatter ().
            reformat (net.grelf.image.FileFormat.TIFF, 32);
        break;

      case "TIFF16":
        new net.grelf.sequence.Reformatter ().
            reformat (net.grelf.image.FileFormat.TIFF, 16);
        break;

      case "TIFF8":
        new net.grelf.sequence.Reformatter ().
            reformat (net.grelf.image.FileFormat.TIFF, 8);
        break;

      case "warpgrid":
        new BatchProcessor (BatchAction.WARPGRID).batch ("_grid", params);
        break;
    }
	} // actionPerformed

  private void slideShow ()
  {
    int interval = net.grelf.Util.askInteger (
            "Interval between slides, s [1..1000]?", 5, 1, 1000);

    if (interval != Integer.MIN_VALUE)
    {
      net.grelf.Util.message (GRIP.getInstance (), "NOTE",
              "In the following file selection dialogue,\n" +
              "if you select only 1 file it will be\n" +
              "taken to be a script file containing\n" +
              "a list of file names (with paths), one\n" +
              "per line.");
      java.io.File [] files = net.grelf.FileIO.selectInputFiles ("Images to show");

      if (null != files && files.length > 0)
      {
        if (files.length == 1)
        {
          new net.grelf.sequence.SlideShow (
                  files [0].getAbsolutePath (), interval * 1000); // ms
        }
        else
        {
          new net.grelf.sequence.SlideShow (files, interval * 1000); // ms
        }
      }
    }
  } // slideShow

	/** Implementing MenuListener. */
	@Override
	public void menuSelected (javax.swing.event.MenuEvent ev)
	{
	} // menuSelected

	/** Implementing MenuListener but does nothing. */
	@Override
	public void menuDeselected (javax.swing.event.MenuEvent ev) {}

	/** Implementing MenuListener but does nothing. */
	@Override
	public void menuCanceled (javax.swing.event.MenuEvent ev) {}

} // BatchMenu
