package net.grelf.grip;

// This Java source file is part of the application GRIP, for processing and
// measuring digital photographs.
// GRIP is available free from www.grelf.net - please always quote this URL in
// connection with GRIP.
//
// Copyright (c) Graham T Relf, UK, 2006
//
// All software contains bugs, even if, we hope, they are only extremely subtle
// ones. The author has tested this software to the best of his abilities to
// ensure that it is generally fit for its purpose of processing and measuring
// digital photographs. The author makes absolutely no claims and gives no
// guarantees whatsoever as to the accuracy of any output from this software.
// Anyone making decisions based on any output from this software has a
// responsibility to first convince themselves of the appropriateness of so
// doing. They should test the software for themselves and understand how it
// works, sufficiently for their needs. Source code and explanatory
// documentation are available for that purpose.

/** Convolution/deconvolution menu (on each ImFrame) and its actions. */
public class ConvolutionMenu
extends javax.swing.JMenu
implements java.awt.event.ActionListener, javax.swing.event.MenuListener
{
	private ImFrame imframe;
//TODO Enable/disable menu options appropriately

	public ConvolutionMenu (ImFrame frame)
	{
		super ("Convolution");
		imframe = frame;
    setMnemonic ('C');
    initialise ();
  } // COnvolutionMenu

  private void initialise ()
  {
    add (net.grelf.Util.menuItem (this,
            "Median filter", "median", 'F', 0));
    add (net.grelf.Util.menuItem (this,
            "Mean filter", "mean", 'E', 0));
    add (net.grelf.Util.menuItem (this,
            "Variance filter", "variance", 'R', 0));
    add (net.grelf.Util.menuItem (this,
            "Nearest extreme filter", "extreme", 'X', 0));
    add (net.grelf.Util.menuItem (this,
            "Rank filter", "rank", 'K', 0));
    add (net.grelf.Util.menuItem (this,
            "Gaussian blur", "blur", 'B', 0));
//    add (net.grelf.Util.menuItem (this,
//            "Unsharp mask", "unsharp", 'U', 0));
    addSeparator ();
    add (net.grelf.Util.menuItem (this,
            "Load kernel from file", "kernelload", 'L', 0));
		add (net.grelf.Util.menuItem (this,
            "Sample kernel manually", "kernelmanual", 'M', 0));
		add (net.grelf.Util.menuItem (this,
            "Sample kernel automatically (star)", "kernelauto", 'A', 0));
		add (net.grelf.Util.menuItem (this,
            "Set Gaussian kernel", "kernelgauss", 'G', 0));
		add (net.grelf.Util.menuItem (this,
            "View kernel", "viewkernel", 'V', 0));
		add (net.grelf.Util.menuItem (this,
            "Convolve", "convolve", 'C', 0));
		addSeparator ();
		add (net.grelf.Util.menuItem (this,
            "Deconvolve (van Cittert)", "deconvolve", 'D', 0));
    addSeparator ();
    add (net.grelf.Util.menuItem (this,
            "Fourier transform (FFT)", "fft", 'T', 0));
    addSeparator ();
/*
    add (net.grelf.Util.menuItem (this,
            "Experiment", "expt", 'P', 0));
		addSeparator ();
*/
    add (net.grelf.Util.menuItem (this,
            "Help about this menu", "help", 'H', 0));
	} // initialise

	/** Implementing the ActionListener interface to act on menu option
    * selection. */
	@Override
	public void actionPerformed (java.awt.event.ActionEvent ev)
	{
		String action = ev.getActionCommand ();

    switch (action)
    {
      case "blur":
        {
          FastSlowDialogue fsd = new FastSlowDialogue (imframe);
          fsd.dispose ();
          switch (fsd.getSelected ())
          {
            case FAST:
              fastBlur ();
              break;

            case SLOW:
              int maxRadius = 16;
              int blurRadius = net.grelf.Util.askInteger (
                      "Blur radius? [1.." + maxRadius + "]", 1, 1, maxRadius);

              if (Integer.MIN_VALUE != blurRadius)
              {
                int width = 1 + 2 * blurRadius;
                imframe.doOpInBackground (OpThread.Ops.BLUR_GAUSSIAN, width);
                imframe.addHistory ("Gaussian blur (" +
                        blurRadius + " x " + blurRadius + ")");
              }

              break;

            case CANCEL:
          }
          break;
        }

      case "convolve":
        {
          FastSlowDialogue fsd = new FastSlowDialogue (imframe);
          fsd.dispose ();
          switch (fsd.getSelected ())
          {
            case FAST:
              fastConvolve ();
              break;

            case SLOW:
              imframe.doOpInBackground (OpThread.Ops.CONVOLVE, 0);
              imframe.addHistory ("Convolution");
              break;

            case CANCEL:
          }
          break;
        }

      case "deconvolve":
        {
          nPasses = net.grelf.Util.askInteger (this,
                  "No of passes? [1..50]", 10, 1, 50);
          deconvolutionWeight = net.grelf.Util.askDouble (this,
                  "Deconvolution weight? [0.1..1.0]", 1.0, 0.1, 1.0);
          FastSlowDialogue fsd = new FastSlowDialogue (imframe);
          fsd.dispose ();
          switch (fsd.getSelected ())
          {
            case FAST:
              fastDeconvolve ();
              break;

            case SLOW:
              imframe.doOpInBackground (OpThread.Ops.DECONVOLVE, 0);
              imframe.addHistory ("Deconvolution");
              break;

            case CANCEL:
          }
          break;
        }

      case "expt":
        net.grelf.image.Image image = imframe.getImage ();
        int wd = image.getWidth ();
        int ht = image.getHeight ();
        net.grelf.image.Image64 image64 =
                (net.grelf.image.Image64) image.convertToImage64 ();
        if (null == image64)
        {
          throw new RuntimeException ("Image not created");
        }

        // Create monochrome Gaussian template, in the 4 corners:
        int halfWidth = net.grelf.Util.askInteger (
                "Half width (px)?", 10, 1, Math.min (wd, ht) / 4);

        if (halfWidth != Integer.MIN_VALUE)
        {
          net.grelf.image.ComplexImage template64fft =
                  image64.createTransformedGaussian (halfWidth * 2 + 1);

          if (null == template64fft)
          {
            throw new RuntimeException ("Image not created");
          }

  //      template64fft.invert ();
  //new ImFrame ("template Im", template64fft.getIm ());
  //new ImFrame ("template Re", template64fft.getRe ());
          net.grelf.image.Image64 im64 =
                  (net.grelf.image.Image64) image.convertToImage64 ();

          if (null == im64)
          {
            return;
          }

          net.grelf.image.ComplexImage cx64 =
                  new net.grelf.image.ComplexImage64 (im64);

          if (null == cx64)
          {
            return;
          }

          net.grelf.image.ComplexImage64 cx64fft =
                  (net.grelf.image.ComplexImage64) cx64.fft ();

          if (null == cx64fft)
          {
            return;
          }

          cx64.dispose ();
          cx64 = null;
  //      cx64fft.multiply (template64fft);
  cx64fft.divide (template64fft);
          template64fft.dispose ();
          template64fft = null;
          cx64 = cx64fft.fftInverse ();
          cx64fft.dispose ();
          cx64fft = null;
  new ImFrame ("result Im", cx64.getImaginary ());
  new ImFrame ("result Re", cx64.getReal ());
        }
        break;

      case "extreme":
        imframe.doOpInBackground (OpThread.Ops.NEAREST_EXTREME);
        break;

      case "fft":
        if (net.grelf.Util.confirm ("FFT",
                "Resulting images have 64 bits per channel - OK?"))
        {
          imframe.doOpInBackground (OpThread.Ops.FFT);
        }

        break;

      case "help":
        HelpMenu.help ("pages/conv_menu.html");
        break;

      case "kernelgauss":
        {
          imframe.addHistory ("Gaussian kernel");
          net.grelf.image.Kernel k = new net.grelf.image.Kernel (0, null);

          if (k.askHalfWidth ())
          {
            setKernel (
                    net.grelf.image.Kernel.createGaussianKernel (
                        imframe.getImage (), k.getHalfWidth ()));
          }

          break;
        }

      case "kernelload":
        String path = net.grelf.FileIO.selectInputFile ("Load kernel CSV file");

        if (null != path)
        {
          imframe.addHistory ("Load kernel from " + path);
          setKernel (
                  net.grelf.image.Kernel.load (
                  path, imframe.getImage ().getNBands ()));
        }

        break;

      case "kernelmanual":
        imframe.addHistory ("Set kernel manually");
        net.grelf.image.Kernel mk = new net.grelf.image.Kernel (0, null);

        if (mk.askHalfWidth ())
        {
          // We need an initial kernel to satisfy GlassPane, for drawing an outline:
          setKernel (mk);
          setInitialHalfWidth (mk.getHalfWidth ());
          imframe.pushCaption ();
          GlassPane gp = imframe.getGP ();
          gp.setDrawingMode (DrawingMode.KERNEL);
          imframe.setTitle ("Click to set kernel centre");
          // The end of the process is then in Kernel ().
        }

        break;

      case "kernelauto":
        imframe.addHistory ("Set kernel automatically");
        // We need an initial kernel to satisfy GlassPane, for drawing an outline:
        net.grelf.image.Kernel ak = new net.grelf.image.Kernel (0, null);
        setKernel (ak);
        setInitialHalfWidth (-50);
        imframe.pushCaption ();
        GlassPane gp = imframe.getGP ();
        gp.setDrawingMode (DrawingMode.KERNEL);
        imframe.setTitle ("Click to set kernel centre");
        // The end of the process is then in Kernel ().
        break;

      case "mean":
        imframe.doOpInBackground (OpThread.Ops.MEAN);
        break;

      case "median":
        imframe.doOpInBackground (OpThread.Ops.MEDIAN);
        break;

      case "rank":
        imframe.doOpInBackground (OpThread.Ops.RANK);
        break;

      case "variance":
        imframe.doOpInBackground (OpThread.Ops.VARIANCE);
        break;

      case "viewkernel":
        {
          net.grelf.image.Kernel k = getKernel ();
          if (null == k)
          {
            net.grelf.Util.warning ("Error", "Kernel is not yet set.");
          }
          else
          {
            viewKernel (imframe.getImage ());
          }
          break;
        }
    }
	} // actionPerformed

  private static int nPasses = 1;
  protected static int getNPasses () { return nPasses; }

  private static double deconvolutionWeight = 0.5;
  protected static double getDeconvolutionWeight () { return deconvolutionWeight; }

	/** Implementing MenuListener. */
	@Override
	public void menuSelected (javax.swing.event.MenuEvent ev) {}

	/** Implementing MenuListener but does nothing. */
	@Override
	public void menuDeselected (javax.swing.event.MenuEvent ev) {}

	/** Implementing MenuListener but does nothing. */
	@Override
	public void menuCanceled (javax.swing.event.MenuEvent ev) {}

	private static net.grelf.image.Kernel kernel;
	public static net.grelf.image.Kernel getKernel () { return kernel; }
	public static void setKernel (net.grelf.image.Kernel aKernel) { kernel = aKernel; }

	private static int initialHalfWidth = 1;
	public static int getInitialHalfWidth () { return initialHalfWidth; }
	/** For the manual and automatic (star) setting of the kernel from an image. */
	public static void setInitialHalfWidth (int aHalfWidth) { initialHalfWidth = aHalfWidth; }

  /** Display the kernel in an ImFrame with the same number of bands and bit depth as the given image. */
	public void viewKernel (net.grelf.image.Image image)
	{
		final int PIXEL_SIZE = 8;
		int kwd = 2 * kernel.getHalfWidth () + 1;
		int wd = kwd * PIXEL_SIZE;
		int ht = kwd * PIXEL_SIZE;
		net.grelf.Util.message ("Note",
		              "Using menu options in the following\n" +
		              "display of the kernel as an image\n" +
		              "will not affect the current kernel.");
		ImFrame imf = new ImFrame (kwd + " x " + kwd + " kernel", wd, ht, true);
		ImPane ip = imf.getImPane ();
    int nb = image.getNBands ();
    net.grelf.image.Image dstImage =
            net.grelf.image.ImageBase.createImage (wd, ht, nb, image.getBitsPerChannel ());
    ip.setImage (dstImage);
    double factor = dstImage.getRangeDouble ().high / image.getRangeDouble ().high;
    double [] px = new double [nb];
    double [][] data = kernel.getData ();

    for (int ypx = 0, k = 0; ypx < ht; ypx += PIXEL_SIZE)
    {
      for (int xpx = 0; xpx < wd; xpx += PIXEL_SIZE, k++)
      {
        for (int band = 0; band < nb; band++)
        {
          px [band] = data [band][k] * factor;
        }

        for (int iy = 0; iy < PIXEL_SIZE; iy++)
        {
          for (int ix = 0; ix < PIXEL_SIZE; ix++)
          {
            dstImage.setPixelDouble (xpx + ix, ypx + iy, px);
          }
        }
      }
    }

		imf.redisplay ();
	} // viewKernel

  private void fastBlur ()
  {
    net.grelf.image.Image image = imframe.getImage ();
    net.grelf.image.Image64 image64 =
            (net.grelf.image.Image64) image.convertToImage64 ();

    if (null == image64)
    {
      throw new RuntimeException ("Image not created");
    }

    int maxRadius = Math.min (image64.getWidth (), image64.getHeight ()) / 4; //Was 16 until 12.8.10
    int blurRadius = net.grelf.Util.askInteger ("Blur radius? [1.." + maxRadius + "]", 1, 1, maxRadius);

    if (Integer.MIN_VALUE != blurRadius)
    {
      net.grelf.image.Timer timer = new net.grelf.image.Timer ("Gaussian blur, radius " + blurRadius, image64);
      imframe.setCursor (GRIP.GRIP_WAIT_CURSOR);

      try
      {
        net.grelf.image.ComplexImage template64fft =
                image64.createTransformedGaussian (blurRadius * 2 + 1);

        if (null == template64fft)
        {
          throw new RuntimeException ("Image not created");
        }

//new ImFrame ("template Im", template64fft.getIm ());
//new ImFrame ("template Re", template64fft.getRe ());

        net.grelf.image.ComplexImage cx64 = new net.grelf.image.ComplexImage64 (image64);

        if (null == cx64)
        {
          throw new RuntimeException ("Image not created");
        }

        net.grelf.image.ComplexImage64 cx64fft = (net.grelf.image.ComplexImage64) cx64.fft ();

        if (null == cx64fft)
        {
          throw new RuntimeException ("Image not created");
        }

        cx64.dispose ();
        cx64 = null;
        cx64fft.multiply (template64fft);
        template64fft.dispose ();
        template64fft = null;
        cx64 = cx64fft.fftInverse ();
        cx64fft.dispose ();
        cx64fft = null;
//new ImFrame ("Gaussian blur, radius " + blurRadius + "px - imaginary part", cx64.getIm ());
        cx64.getImaginary ().dispose ();
//new ImFrame ("Gaussian blur, radius " + blurRadius + "px - real part", cx64.getReal ());

        int bits = image.getBitsPerChannel ();
        image.dispose ();

        switch (bits)
        {
          case  8:
            image = cx64.getReal ().convertToImage8 ();
            cx64.getReal ().dispose ();
            break;

          case 16:
            image = cx64.getReal ().convertToImage16 ();
            cx64.getReal ().dispose ();
            break;

          case 32:
            image = cx64.getReal ().convertToImage32 ();
            cx64.getReal ().dispose ();
            break;

          default:
            image = cx64.getReal ();
        }

        if (null == image)
        {
          throw new RuntimeException ("Image not created");
        }

        imframe.getImPane ().setImage (image);
        imframe.redisplay ();
        imframe.addHistory ("Gaussian blur, radius " + blurRadius + "px");
        timer.stop ();
      }
      catch (RuntimeException ex)
      {
        net.grelf.Util.warning (imframe, "Error", "Gaussian blur aborted");
      }
      finally
      {
        imframe.setCursor (imframe.IMFRAME_DEFAULT_CURSOR);
      }
    }
  } // fastBlur

  private void fastConvolve ()
  {
    imframe.setCursor (GRIP.GRIP_WAIT_CURSOR);
    net.grelf.image.Image image = imframe.getImage ();
    int wd = image.getWidth ();
    int ht = image.getHeight ();
    net.grelf.image.Image64 image64 =
            (net.grelf.image.Image64) image.convertToImage64 ();

    if (null == image64)
    {
      throw new RuntimeException ("Image not created");
    }

    try
    {
      net.grelf.image.Timer timer = new net.grelf.image.Timer ("Convolution", image64);
      net.grelf.image.ComplexImage kx64fft =
              image64.createTransformedKernel (kernel);

      if (null == kx64fft)
      {
        throw new RuntimeException ("kx64fft not created");
      }

      net.grelf.image.ComplexImage64 ix64 = new net.grelf.image.ComplexImage64 (image64);

      if (null == ix64)
      {
        throw new RuntimeException ("ix64 not created");
      }

      net.grelf.image.ComplexImage64 ix64fft = (net.grelf.image.ComplexImage64) ix64.fft ();

      if (null == ix64fft)
      {
        throw new RuntimeException ("ix64fft not created");
      }

      ix64fft.multiply (kx64fft);
      net.grelf.image.ComplexImage64 kix64 = (net.grelf.image.ComplexImage64) ix64fft.fftInverse ();
      ix64fft.dispose ();

      if (null == kix64)
      {
        throw new RuntimeException ("kix64 not created");
      }

      imframe.getImPane ().setImage (kix64.getReal ());
      imframe.redisplay ();
      net.grelf.Util.message (imframe, "Info", "Convolution done");
      imframe.addHistory ("Convolution");
      timer.stop ();
    }
    catch (RuntimeException ex)
    {
      net.grelf.Util.warning ("Error", "Convolution aborted:\n" + ex.getMessage ());
    }
    finally
    {
      imframe.setCursor (imframe.IMFRAME_DEFAULT_CURSOR);
    }
  } // fastConvolve

  private void fastDeconvolve ()
  {
    imframe.setCursor (GRIP.GRIP_WAIT_CURSOR);
    net.grelf.image.Image image = imframe.getImage ();
    int wd = image.getWidth ();
    int ht = image.getHeight ();
    net.grelf.image.Image64 image64 =
            (net.grelf.image.Image64) image.convertToImage64 ();

    if (null == image64)
    {
      throw new RuntimeException ("Image not created");
    }

    net.grelf.image.Timer timer = new net.grelf.image.Timer ("Deconvolution", image64);

    // Create monochrome Gaussian template, in the 4 corners:
    int halfWidth = net.grelf.Util.askInteger ("Half width (px)?", 10, 1, Math.min (wd, ht) / 4);

    if (halfWidth != Integer.MIN_VALUE)
    {
/* Method (x in name means ComplexImage):
Given original image ix64 (stays unaltered)
and Gaussian kernel, k46

kx64fft = k64.fft ();
k64.dispose ();
ox64 = ix64.clone ();

for (passes)
{
ox64fft = ox64.fft ();
ox64fft = ox64fft.multiply (kx64fft);
kox64 = ox64fft.fftInverse ();
ox64fft.dispose ();
cx64 = ix64.clone ();
cx64.subtractSimple (kox64);
cx64.multiply (w);
ox64 = ox64.add (cx64);
cx64.dispose ();
}
*/
      try
      {
        net.grelf.image.ComplexImage kx64fft =
                image64.createTransformedGaussian (halfWidth * 2 + 1);

        if (null == kx64fft)
        {
          throw new RuntimeException ("kx64fft not created");
        }

        net.grelf.image.Image64 im64 = (net.grelf.image.Image64) image.convertToImage64 ();

        if (null == im64)
        {
          throw new RuntimeException ("im64 not created");
        }

        net.grelf.image.ComplexImage64 ix64 = new net.grelf.image.ComplexImage64 (im64);
        net.grelf.image.ComplexImage64 ox64 =
                new net.grelf.image.ComplexImage64 ((net.grelf.image.Image64) im64.clone ());

        if (null == ix64 || null == ox64)
        {
          throw new RuntimeException ("ix64 or ox64 not created");
        }

        for (int pass = 1; pass <= nPasses; pass++)
        {
          net.grelf.image.ComplexImage64 ox64fft = (net.grelf.image.ComplexImage64) ox64.fft ();

          if (null == ox64fft)
          {
            throw new RuntimeException ("ox64fft not created");
          }

          ox64fft.multiply (kx64fft);
          net.grelf.image.ComplexImage64 kox64 = (net.grelf.image.ComplexImage64) ox64fft.fftInverse ();
          ox64fft.dispose ();

          if (null == kox64)
          {
            throw new RuntimeException ("kox64 not created");
          }

          net.grelf.image.ComplexImage64 cx64 =
                new net.grelf.image.ComplexImage64 ((net.grelf.image.Image64) im64.clone ());

          cx64.subtract (kox64); // correction term
          cx64.getReal ().multiplyWeighted (deconvolutionWeight);
          cx64.getImaginary ().multiplyWeighted (deconvolutionWeight);
          ox64.add (cx64); // o'
          cx64.dispose ();
        }

        imframe.getImPane ().setImage (ox64.getReal ());
        imframe.redisplay ();
        net.grelf.Util.message (imframe, "Info", "Deconvolution done");
        imframe.addHistory ("Deconvolution");
        timer.stop ();
      }
      catch (net.grelf.image.IncompatibleImageException ex)
      {
        net.grelf.Util.warning ("Error", ex.toString () + "\n" + ex.getMessage ());
      }
      catch (RuntimeException ex)
      {
        net.grelf.Util.warning ("Error", "Deconvolution aborted:\n" + ex.getMessage ());
      }
      finally
      {
        imframe.setCursor (imframe.IMFRAME_DEFAULT_CURSOR);
      }
    }
  } // fastDeconvolve

} // ConvolutionMenu