package net.grelf.grip;

// This Java source file is part of the application GRIP, for processing and
// measuring digital photographs.
// GRIP is available free from www.grelf.net - please always quote this URL in
// connection with GRIP.
//
// Copyright (c) Graham T Relf, UK, 2006
//
// All software contains bugs, even if, we hope, they are only extremely subtle
// ones. The author has tested this software to the best of his abilities to
// ensure that it is generally fit for its purpose of processing and measuring
// digital photographs. The author makes absolutely no claims and gives no
// guarantees whatsoever as to the accuracy of any output from this software.
// Anyone making decisions based on any output from this software has a
// responsibility to first convince themselves of the appropriateness of so
// doing. They should test the software for themselves and understand how it
// works, sufficiently for their needs. Source code and explanatory
// documentation are available for that purpose.

/** Help menu and its actions. Needs Java 6 for the browser launching
  * capabilities of new java.awt.Desktop. */
public class HelpMenu
extends javax.swing.JMenu
implements java.awt.event.ActionListener, javax.swing.event.MenuListener
{

	public HelpMenu ()
	{
		super ("Help");
    setMnemonic ('H');
    initialise ();
  } // HelpMenu

  private void initialise ()
  {
		add (net.grelf.Util.menuItem (this,
            "Help contents", "helpcontents",
            'H', java.awt.event.KeyEvent.VK_F1, false)); // No Ctrl
		add (net.grelf.Util.menuItem (this,
            "Help index", "helpindex",
            'I', 0));
		add (net.grelf.Util.menuItem (this,
            "API documentation (for programmers)", "helpapi",
            'D', 0));
		add (net.grelf.Util.menuItem (this,
            "About GRIP", "helpabout",
            'G', 0));
	} // initialise

	/** Implementing ActionLister, to respond to menu option selection. */
	@Override
	public void actionPerformed (java.awt.event.ActionEvent ev)
	{
		String action = ev.getActionCommand ();

    switch (action)
    {
      case "helpabout":
        GRIP.about ();
        break;

      case "helpapi":
        help ("docs/index.html");
        break;

      case "helpcontents":
        help ("pages/contents.html");
        break;

      case "helpindex":
        help ("pages/idx.html");
        break;
    }
	} // actionPerformed

	/** Open a specified local file in the user's default browser.
    * Requires Java 6 or later because it uses java.awt.Desktop.
		* <br>
		* 9.7.27: Made public and static. */
	public static void help (String path)
	{
		if (java.awt.Desktop.isDesktopSupported ())
		{
			java.awt.Desktop d = java.awt.Desktop.getDesktop ();

//			if (d.isSupported (java.awt.Desktop.Action.BROWSE))
// Doesn't seem to work as advertised in Java 6 API
			if (d.isSupported (java.awt.Desktop.Action.OPEN))
			{
				try
				{
					String filePath;
					int idx = path.indexOf ('#');

					if (-1 != idx)
					{
						filePath = path.substring (0, idx);
					}
					else
					{
						filePath = path;
					}

					if (!(new java.io.File (filePath)).exists ())
					{
						net.grelf.Util.warning ("Sorry",
                    "File " + filePath + " does not exist.");
					}
					else
					{
//					d.browse (new java.net.URI (path));
						d.open (new java.io.File (filePath));
					}
        }
        catch (java.io.IOException ex)
        {
					net.grelf.Util.warning ("IOException", "Could not launch browser");
        }
//        catch (java.net.URISyntaxException ex)
//        {
//					net.grelf.Util.warning ("URISyntaxException",
//              "Could not launch browser");
//        }
			}
			else
			{
				net.grelf.Util.warning ("Sorry",
                "Browser launching is not supported on this platform");
			}
		}
		else
		{
			net.grelf.Util.warning ("Sorry",
              "Desktop capabilities are not supported on this platform");
		}
	} // help

	/** Implementing MenuListener. */
	@Override
	public void menuSelected (javax.swing.event.MenuEvent ev)
	{
	} // menuSelected

	/** Implementing MenuListener but does nothing. */
	@Override
	public void menuDeselected (javax.swing.event.MenuEvent ev) {}

	/** Implementing MenuListener but does nothing. */
	@Override
	public void menuCanceled (javax.swing.event.MenuEvent ev) {}

} // HelpMenu
