package net.grelf.grip;

// This Java source file is part of the application GRIP, for processing and
// measuring digital photographs.
// GRIP is available free from www.grelf.net - please always quote this URL in
// connection with GRIP.
//
// Copyright (c) Graham T Relf, UK, 2006
//
// All software contains bugs, even if, we hope, they are only extremely subtle
// ones. The author has tested this software to the best of his abilities to
// ensure that it is generally fit for its purpose of processing and measuring
// digital photographs. The author makes absolutely no claims and gives no
// guarantees whatsoever as to the accuracy of any output from this software.
// Anyone making decisions based on any output from this software has a
// responsibility to first convince themselves of the appropriateness of so
// doing. They should test the software for themselves and understand how it
// works, sufficiently for their needs. Source code and explanatory
// documentation are available for that purpose.

/** Measurement menu (on each ImFrame) and its actions. */
public class MeasurementMenu
extends javax.swing.JMenu
implements java.awt.event.ActionListener, javax.swing.event.MenuListener
{
	private ImFrame imframe;
	private javax.swing.JMenuItem itemBlob, itemReview, itemCalibStar, itemSpectrum,
                                item2DProf, item3DHist /*, itemControl */;

	public MeasurementMenu (ImFrame frame)
	{
		super ("Measure");
		imframe = frame;
    setMnemonic ('M');
    initialise ();
  } // MeasurementMenu

  private void initialise ()
  {
		add (net.grelf.Util.menuItem (this,
            "Hover and magnify", "meashover",
            'H', java.awt.event.KeyEvent.VK_H));
		add (net.grelf.Util.menuItem (this,
            "Single point data", "measpoint",
            'P', 0));
		add (itemBlob = net.grelf.Util.menuItem (this,
            "Blobs/stars separation data", "blobpair",
            'B', java.awt.event.KeyEvent.VK_B));
		add (net.grelf.Util.menuItem (this,
            "Straight line profile", "measline", 'L', 0));
		add (net.grelf.Util.menuItem (this,
            "Curve (freehand) profile", "meascurve", 'C', 0));
		add (net.grelf.Util.menuItem (this,
            "Area (freehand) histogram", "measarea", 'A', 0));
		add (net.grelf.Util.menuItem (this,
            "Rectangle histogram", "measrect", 'R', 0));
		add (net.grelf.Util.menuItem (this,
            "Whole image histogram", "measwhole", 'W', 0));
		add (net.grelf.Util.menuItem (this,
            "Same shape as...", "measshape", 'S', 0));
		add (net.grelf.Util.menuItem (this,
            "Move shape", "moveshape", 'M', 0));
		addSeparator ();
		add (itemReview = net.grelf.Util.menuItem (this,
            "Review last measurements", "measshow", 'E', 0));
		add (net.grelf.Util.menuItem (this,
            "Compare open images", "compare", 'I', 0));
		addSeparator ();
		add (net.grelf.Util.menuItem (this,
            "Calibrate distance", "calibdist", 'D', 0));
		add (itemCalibStar = net.grelf.Util.menuItem (this,
            "Calibrate on known stars", "calibstar", 'K', 0));
		add (net.grelf.Util.menuItem (this,
            "View calibration", "viewcalib", 'V', 0));
		add (net.grelf.Util.menuItem (this,
            "Same calibration as...", "copycalib", 'N', 0));
		addSeparator ();
		add (net.grelf.Util.menuItem (this,
            "Clear overlay", "clear", 'O', 0));
		add (net.grelf.Util.menuItem (this,
            "Draw overlay into new image", "draw", 'Y', 0));
		addSeparator ();
		add (itemSpectrum = net.grelf.Util.menuItem (this,
            "Star spectrum", "spectrum", 'T', 0));
		add (item2DProf = net.grelf.Util.menuItem (this,
            "2D profile", "2dprof", '2', 0));
		add (item3DHist = net.grelf.Util.menuItem (this,
            "3D histogram", "3dhist", '3', 0));
		addSeparator ();
//		add (itemControl = net.grelf.Util.menuItem (this,
//          "Set multiple control points", "control", 'U', 0));
//		addSeparator ();
		add (net.grelf.Util.menuItem (this,
            "Help about this menu", "help", 0, 0));

		addMenuListener (this);
    // Need to be able to revise settings when the menu is invoked
	} // iniialise

	/** Enable/disable menu options as necessary. */
	public void reviseSettings ()
	{
		this.itemBlob.setEnabled (null != this.imframe.getBlobMeases ()
            || (this.imframe instanceof StarChartFrame
                && null != ((StarChartFrame) this.imframe).getStarChart ()));

    itemSpectrum.setEnabled (true);
    item2DProf.setEnabled (true);
    item3DHist.setEnabled (true);
//      itemControl.setEnabled (true);

		GlassPane gp = imframe.getGP ();

		switch (gp.getDrawingMode ())
		{
		case CLOSED_CURVE:
		case OPEN_CURVE:
		case POINT:
		case POINT_PAIR:
		case RECTANGLE:
		case WHOLE:
			this.itemReview.setEnabled (true);
			break;

		default:
			this.itemReview.setEnabled (false);
		}

		int nStarsWithPosition = 0;
		BlobMeasList meases = this.imframe.getBlobMeases ();

		if (null != meases)
		{
			for (BlobMeas meas : meases)
			{
				net.grelf.astro.Star star = meas.getStar ();

				if (null != star && null != star.getPosition ())
				{
					nStarsWithPosition++;
				}
			}
		}

		this.itemCalibStar.setEnabled (nStarsWithPosition > 1);
	} // reviseSettings

	/** Implementing the ActionListener interface to act on menu option
    * selection. */
	@Override
	public void actionPerformed (java.awt.event.ActionEvent ev)
	{
		String action = ev.getActionCommand ();

    switch (action)
    {
      case "2dprof":
        Prof3D_Pane.display (imframe);
        break;

      case "3dhist":
        if (3 != imframe.getImage ().getNBands ())
        {
          net.grelf.Util.message ("Sorry",
                  "3D histogram is only for RGB images");
        }
        else
        {
          Hist3D_Pane.display (imframe);
        }

        break;

      case "calibdist":
        imframe.addHistory ("Calibrate distance");
        imframe.getGP ().setMeasured (false);
        Calibrator.calibrate (imframe);
        break;

      case "calibstar":
        imframe.addHistory ("Calibrated on stars");
        Calibrator.calibrateOnStars (imframe);
        break;

      case "clear":
        clearOverlay (imframe);
        break;

      case "compare":
        imframe.addHistory ("Compare measurements");
        new MeasurementTable (imframe);
        break;

      case "copycalib":
        copyCalibrationFromOtherFrame (imframe);
        break;

      case "draw":
        imframe.drawOverlay (true); // New image
        break;

      case "help":
        HelpMenu.help ("pages/meas_menu.html");
        break;

      case "blobpair":
        measureBetweenBlobs (imframe);
        break;

      case "measarea":
        measureArea (imframe);
        break;

      case "meascurve":
        measureCurve (imframe);
        break;

      case "meashover":
        measureHovering (imframe);
        break;

      case "measline":
        measureStraightLine (imframe);
        break;

      case "measpoint":
        measurePoint (imframe);
        break;

      case "measrect":
        measureRectangle (imframe);
        break;

      case "measshape":
        measureSameShapeAsOtherFrame (imframe);
        break;

      case "measshow":
        reviewLastMeasurements (imframe);
        break;

      case "measwhole":
        measureWholeImage (imframe);
        break;

      case "moveshape":
        moveShape (imframe);
        break;

      case "spectrum":
        imframe.analyseAsSpectrum ();
        break;

      case "viewcalib":
        Calibrator.show (imframe);
        break;
    }
	} // actionPerformed

	/** Erase the drawn shape or other overlay from the GlassPane of the
    * given frame. */
	public void clearOverlay (ImFrame imframe)
	{
			imframe.addHistory ("Clear overlay");
			imframe.getGP ().setDrawingMode (DrawingMode.UNDEFINED);
			imframe.popCaption ();
	} // clearOverlay

	/** Ask user to select another calibrated frame from which calibration
    * will be copied to the given one. */
	public void copyCalibrationFromOtherFrame (ImFrame imframe)
	{
		java.util.List<ImFrame> frames = GRIP.getImFramesList ();
		java.util.List<String> captions = new java.util.ArrayList <> ();

		for (ImFrame frame : frames)
		{
			if (frame != imframe)
			{
				if (frame.getImage ().getCalibration ().isCalibrated ())
				{
					captions.add (frame.getCaption ());
				}
			}
		}

		ImFrame otherFrame = null;

		if (0 == captions.size ())
		{
			net.grelf.Util.warning ("Sorry", "No other frames have been calibrated");
		}
		else
		if (1 == captions.size ())
		{
			for (int i = 0; i < frames.size (); i++)
			{
				if (captions.get (0).equals (frames.get (i).getCaption ()))
				{
					if (net.grelf.Util.confirm (imframe,
                  "Use same calibration as...",
                  "...set in " + captions.get (0) + "?"))
					{
						otherFrame = frames.get (i);
					}

					break;
				}
			}
		}
		else
		{
			String [] dropDown = new String [captions.size ()];

			for (int i = 0; i < captions.size (); i++)
			{
				dropDown [i] = captions.get (i);
			}

			String selected = (String) javax.swing.JOptionPane.showInputDialog (
              imframe, "Use same calibration as...",
							"Choose an image frame",
							javax.swing.JOptionPane.QUESTION_MESSAGE,
							null, dropDown, dropDown [0]);

			if (null != selected)
			{
				for (int i = 0; i < frames.size (); i++)
				{
					if (selected.equals (frames.get (i).getCaption ()))
					{
						otherFrame = frames.get (i);
						break;
					}
				}
			}
		}

		if (null != otherFrame)
		{
			imframe.addHistory ("Copy calibration");
			imframe.getImage ().setCalibration (
                    otherFrame.getImage ().getCalibration ());
		}
	} // copyCalibrationFromOtherFrame

	/** The user defines a closed curve on the given frame, which is then
    * measured. */
	public void measureArea (ImFrame imframe)
	{
		imframe.addHistory ("Measure freehand area");
		imframe.getGP ().setMeasured (false);
		imframe.pushCaption ();
		GlassPane gp = imframe.getGP ();
		gp.setDrawingMode (DrawingMode.CLOSED_CURVE);
		imframe.setTitle ("Drag mouse to draw outline, auto-closes at end");
		// Interaction is then in ImPane.
		// The end of the measurement process is then in PointList.measure ().
	} // measureArea

	/** The user defines two points which are the centres of the blobs nearest
    * to the mouse cursor. The straight line between blob centres is then
    * measured. If both blobs have been identified as stars, with
		* net.grelf.astro.SkyPoints, the inter-star distance is shown as a
    * net.grelf.astro.SkyVector (PA and angular separation) as well as
    * measurements within the image itself. */
	public void measureBetweenBlobs (ImFrame imframe)
	{
		imframe.addHistory ("Measure between blobs/stars");
		imframe.getGP ().setMeasured (false);
		imframe.pushCaption ();
		GlassPane gp = imframe.getGP ();

		if (null != imframe.getBlobMeases ())
		{
			gp.setDrawingMode (DrawingMode.BLOB_PAIR);
		}
		else
		if (imframe instanceof StarChartFrame
     && null != ((StarChartFrame) imframe).getStarChart ())
		{
			gp.setDrawingMode (DrawingMode.STAR_PAIR);
		}
		else
		{
			net.grelf.Util.logWarning (
"User was able to select Measure between blobs/stars but could not find blobs or stars");
			return; // Should never occur
		}

		imframe.setTitle ("Click to set line ends on nearest blobs/stars");
		// Interaction is then in ImPane.
		// The end of the measurement process is then in PointList.measure ().
	} // measureBetweenBlobs

	/** The user defines an open curve on the given frame, which is then
    * measured. */
	public void measureCurve (ImFrame imframe)
	{
		imframe.addHistory ("Measure freehand curve");
		imframe.getGP ().setMeasured (false);
		imframe.pushCaption ();
		GlassPane gp = imframe.getGP ();
		gp.setDrawingMode (DrawingMode.OPEN_CURVE);
		imframe.setTitle ("Drag mouse to draw curve");
		// Interaction is then in ImPane.
		// The end of the measurement process is then in PointList.measure ().
	} // measureCurve

	/** Set HOVER drawing mode on the given frame so the user can read off
    * measurements. */
	public void measureHovering (ImFrame imframe)
	{
		imframe.getGP ().setMeasured (false);
		imframe.pushCaption ();
		GlassPane gp = imframe.getGP ();
		gp.setDrawingMode (DrawingMode.HOVER);
		imframe.setTitle ("Mouse over image to see values. Click magnifies");
		// Interaction is then in ImPane.
	} // measureHovering

	/** The user defines a single point on the given frame, which is then
    * measured. */
	public void measurePoint (ImFrame imframe)
	{
		imframe.addHistory ("Measure single point");
		imframe.getGP ().setMeasured (false);
		imframe.pushCaption ();
		GlassPane gp = imframe.getGP ();
		gp.setDrawingMode (DrawingMode.POINT);
		imframe.setTitle ("Click to set a single point");
		// Interaction is then in ImPane.
		// The end of the measurement process is then in PointList.measure ().
	} // measurePoint

	/** The user defines a rectangle on the given frame, which is then measured. */
	public void measureRectangle (ImFrame imframe)
	{
		imframe.addHistory ("Measure rectangle");
		imframe.getGP ().setMeasured (false);
		imframe.pushCaption ();
		GlassPane gp = imframe.getGP ();
		gp.setDrawingMode (DrawingMode.RECTANGLE);
		imframe.setTitle ("Click to set rectangle corners (Ctrl for square)");
		// Interaction is then in ImPane.
		// The end of the measurement process is then in PointList.measure ().
	} // measureRectangle

	/** Ask the user to select another measured frame and for its shape then to
    * be measured in the given frame. */
	public void measureSameShapeAsOtherFrame (ImFrame imframe)
	{
		java.util.List<ImFrame> frames = GRIP.getImFramesList ();
		java.util.List<String> captions = new java.util.ArrayList <> ();

		for (ImFrame frame : frames)
		{
			if (frame != imframe)
			{
				GlassPane gp = frame.getGP ();

				if (null != gp.getPoints ())
				{
					captions.add (frame.getCaption ());
				}
			}
		}

		ImFrame otherFrame = null;

		if (0 == captions.size ())
		{
			net.grelf.Util.warning ("Sorry", "No other frames have got measurements");
		}
		else
		if (1 == captions.size ())
		{
			for (int i = 0; i < frames.size (); i++)
			{
				if (captions.get (0).equals (frames.get (i).getCaption ()))
				{
					if (net.grelf.Util.confirm (imframe,
                  "Measure same shape as...", "...measured in " +
                  captions.get (0) + "?"))
					{
						otherFrame = frames.get (i);
					}

					break;
				}
			}
		}
		else
		{
			String [] dropDown = new String [captions.size ()];

			for (int i = 0; i < captions.size (); i++)
			{
				dropDown [i] = captions.get (i);
			}

			String selected = (String) javax.swing.JOptionPane.showInputDialog (
              imframe, "Measure same shape as...", "Choose an image frame",
						  javax.swing.JOptionPane.QUESTION_MESSAGE, null,
              dropDown, dropDown [0]);

			if (null != selected)
			{
				for (int i = 0; i < frames.size (); i++)
				{
					if (selected.equals (frames.get (i).getCaption ()))
					{
						otherFrame = frames.get (i);
						break;
					}
				}
			}
		}

		if (null != otherFrame)
		{
			GlassPane otherGP = otherFrame.getGP ();
			DrawingMode otherDM = otherGP.getDrawingMode ();
			PointList otherPS = otherGP.getPoints ();
			GlassPane thisGP = imframe.getGP ();
			thisGP.setDrawingMode (otherDM);
			thisGP.setPoints (new PointList (otherPS));
			PointList thisPS = thisGP.getPoints ();

			switch (otherDM)
			{
			case POINT:
				thisPS.measurePoint (imframe);
				break;

			case POINT_PAIR:
				thisPS.measureLine (imframe);
				break;

			case RECTANGLE:
				thisPS.measureRectangle (imframe, DrawingMode.RECTANGLE);
				break;

			case OPEN_CURVE:
				thisPS.measureCurve (imframe);
				break;

			case CLOSED_CURVE:
				thisPS.measureArea (imframe);
				break;

			case WHOLE:
				thisPS.measureWhole (imframe);
			}
		}
	} // measureSameShapeAsOtherFrame

	/** The user defines a straight line on the given frame, which is then
    * measured. */
	public void measureStraightLine (ImFrame imframe)
	{
		imframe.addHistory ("Measure straight line");
		imframe.getGP ().setMeasured (false);
		imframe.pushCaption ();
		GlassPane gp = imframe.getGP ();
		gp.setDrawingMode (DrawingMode.POINT_PAIR);
		imframe.setTitle ("Click to set line end points");
		// Interaction is then in ImPane.
		// The end of the measurement process is then in PointList.measure ().
	} // measureStraightLine

	/** Measure the whole image in the given frame and display the results. */
	public void measureWholeImage (ImFrame imframe)
	{
		imframe.addHistory ("Measure whole image");
		imframe.getGP ().setMeasured (false);
		imframe.pushCaption ();
		GlassPane gp = imframe.getGP ();
		gp.setDrawingMode (DrawingMode.WHOLE);
		imframe.setTitle ("Measuring whole image");
		gp.getPoints ().measureWhole (imframe);
		imframe.popCaption ();
		gp.setMeasured (true);
	} // measureWholeImage

	/** Allow the user to drag the shape in the given frame, which is
    * remeasured on releasing the mouse. */
	public void moveShape (ImFrame imframe)
	{
		GlassPane gp = imframe.getGP ();
		PointList pts = gp.getPoints ();

		if (null != pts && 0 < pts.size ())
		{
			imframe.addHistory ("Move shape");
			gp.setMeasured (false);
			imframe.pushCaption ();
			gp.pushDrawingMode ();
			gp.setDrawingMode (DrawingMode.MOVE);
			imframe.setTitle ("Click and drag shape; release to measure");
			// Interaction is then in ImPane.
			// The end of the measurement process is then in PointList.measure ().
		}
		else
		{
			net.grelf.Util.warning ("Sorry",
              "There is no measuring shape you could move");
		}
	} // moveShape

	/** Redisplay the last set of measurements made on the given frame. */
	public void reviewLastMeasurements (ImFrame imframe)
	{
		GlassPane gp = imframe.getGP ();
		PointList ps = gp.getPoints ();

		if (ps != null)
		{
			switch (gp.getDrawingMode ())
			{
			case BLOB_PAIR:
				ps.measureBlobLine (imframe);
				break;

			case CLOSED_CURVE:
				ps.measureArea (imframe);
				break;

			case OPEN_CURVE:
				ps.measureCurve (imframe);
				break;

			case POINT:
				ps.measurePoint (imframe);
				break;

			case POINT_PAIR:
				ps.measureLine (imframe);
				break;

			case RECTANGLE:
				ps.measureRectangle (imframe, DrawingMode.RECTANGLE);
				break;

			case STAR_PAIR:
				ps.measureStarLine (imframe);
				break;

			case WHOLE:
				ps.measureWhole (imframe);
			}
		}
	} // reviewLastMeasurements

/*
	public void setControlPoints ()
	{
		GlassPane gp = imframe.getImPane ().getGlassPane ();
		gp.setDrawingMode (DrawingMode.CONTROL_POINTS);
		imframe.setTitle ("Add control points");
		// Interaction is then in ImPane.
	}
*/

	/** Calls reviseSettings (). Implementing MenuListener. */
	@Override
	public void menuSelected (javax.swing.event.MenuEvent ev)
	{
		reviseSettings ();
	} // menuSelected

	/** Implementing MenuListener but does nothing. */
	@Override
	public void menuDeselected (javax.swing.event.MenuEvent ev) {}

	/** Implementing MenuListener but does nothing. */
	@Override
	public void menuCanceled (javax.swing.event.MenuEvent ev) {}

} // MeasurementMenu