package net.grelf.grip;

// This Java source file is part of the application GRIP, for processing and
// measuring digital photographs.
// GRIP is available free from www.grelf.net - please always quote this URL in
// connection with GRIP.
//
// Copyright (c) Graham T Relf, UK, 2006
//
// All software contains bugs, even if, we hope, they are only extremely subtle
// ones. The author has tested this software to the best of his abilities to
// ensure that it is generally fit for its purpose of processing and measuring
// digital photographs. The author makes absolutely no claims and gives no
// guarantees whatsoever as to the accuracy of any output from this software.
// Anyone making decisions based on any output from this software has a
// responsibility to first convince themselves of the appropriateness of so
// doing. They should test the software for themselves and understand how it
// works, sufficiently for their needs. Source code and explanatory
// documentation are available for that purpose.

/** Segmentation menu (on each ImFrame) and its actions. */
public class SegmentationMenu
extends javax.swing.JMenu
implements java.awt.event.ActionListener, javax.swing.event.MenuListener
{
	private ImFrame imframe;
	private javax.swing.JMenuItem itemBlobMeas, itemBlobBright, itemBlobDrawB,
       itemBlobDrawR, itemBlobHover, itemBlobSave, itemAuto, itemWarping,
       itemClearOverlay, itemDrawOverlay, itemSplit;

	/** To enable menu options for detected blobs, give the menu a reference to the
    * list of blob measurements. */
	public void setBlobMeases (BlobMeasList meases) { this.blobMeases = meases; }
	private BlobMeasList blobMeases = null;

	private Pseudocolour lut = null;

	/** Enable/disable menu options according to the number of channels and bits
    * per channel in the currently displayed image and whether blob detection
    * has been done. */
	public void reviseSettings ()
	{
		if (null == blobMeases)
		{
			itemBlobMeas.setEnabled (false);
			itemBlobBright.setEnabled (false);
			itemBlobDrawB.setEnabled (false);
			itemBlobDrawR.setEnabled (false);
			itemBlobHover.setEnabled (false);
			itemBlobSave.setEnabled (false);
      itemWarping.setEnabled (false);
      itemClearOverlay.setEnabled (false);
      itemDrawOverlay.setEnabled (false);
		}
		else
		{
			itemBlobMeas.setEnabled (true);
			itemBlobBright.setEnabled (true);
			itemBlobDrawB.setEnabled (true);
			itemBlobDrawR.setEnabled (true);
			itemBlobHover.setEnabled (true);
			itemBlobSave.setEnabled (true);
      itemWarping.setEnabled (true);
      itemClearOverlay.setEnabled (true);
      itemDrawOverlay.setEnabled (true);
		}
	} // reviseSettings

	public SegmentationMenu (ImFrame frame)
	{
		super ("Segmentation");
		this.imframe = frame;
    setMnemonic ('S');
    initialise ();
  } // SegmentationMenu

  private void initialise ()
  {
		add (net.grelf.Util.menuItem (
            this, "Threshold", "threshold", 'T', 0));
		add (net.grelf.Util.menuItem (
            this, "Detect blobs", "blobs", 'D', 0));
		add (itemAuto = net.grelf.Util.menuItem (
            this, "Auto-segment (stars)", "autosegment", 'A',
                                                java.awt.event.KeyEvent.VK_E));
    itemAuto.setToolTipText ("Assumes bright stars on dark background");
		add (net.grelf.Util.menuItem (
            this, "Auto-segment (non-saturated stars)", "autosegnonsat", 'N', 0));
    add (itemBlobMeas = net.grelf.Util.menuItem (
            this, "Blob/star measurements", "blobmeas", 'M', 0));
		add (itemBlobBright = net.grelf.Util.menuItem (
            this, "Show brightest", "blobbright", 'W', 0));
		add (itemBlobDrawB = net.grelf.Util.menuItem (
            this, "Overlay boundaries", "blobdrawB", 'B', 0));
		add (itemBlobDrawR = net.grelf.Util.menuItem (
            this, "Overlay regions", "blobdrawR", 'R', 0));
		add (itemClearOverlay = net.grelf.Util.menuItem (
            this, "Clear overlay", "clear", 'C', 0));
		add (itemDrawOverlay = net.grelf.Util.menuItem (
            this, "Draw overlay into new image", "drawoverlay", 'O', 0));
		add (itemBlobHover = net.grelf.Util.menuItem (
            this, "Hover to view blobs/stars", "blobhover", 'H', 0));
		add (itemBlobSave = net.grelf.Util.menuItem (
            this, "Save all blob/star data", "blobsave", 'S', 0));
		add (itemWarping = net.grelf.Util.menuItem (
            this, "Make warping grid", "warping", 'G', 0));
    add (itemSplit = net.grelf.Util.menuItem (
            this, "Split stars from background", "split", 'P', 0));
		addSeparator ();
		add (net.grelf.Util.menuItem (this, "Help about this menu", "help", 'E', 0));
		reviseSettings ();

		addMenuListener (this); // Need to be able to revise settings when the menu is invoked
	} // initialise

	/** Implementing the ActionListener interface to act on menu option selection. */
	@Override
	public void actionPerformed (java.awt.event.ActionEvent ev)
	{
		String action = ev.getActionCommand ();

    switch (action)
    {
      case "autosegment":
        try
        {
          StarSegmenter.segment (imframe, Integer.parseInt (
                  Config.getValue ("StarSegmentDifference.value")),
                                          Integer.parseInt (
                  Config.getValue ("StarSegmentRadius.value")),
                  true); // Erode/dilate
          imframe.addHistory ("Segment stars automatically");
          boolean includeSinglePixelBlobs = false;
          imframe.setCursor (GRIP.GRIP_WAIT_CURSOR);
          blobMeases = imframe.detectBlobs (includeSinglePixelBlobs);

          if (null != blobMeases && blobMeases.size () > 0)
          {
            imframe.setBlobMeases (blobMeases);
            // Set reference in ImFrame so that measurements can be shown.

            reviseSettings ();
            imframe.getGP ().setDrawingMode (DrawingMode.BLOB_HOVER);
          }
          else
          {
            net.grelf.Util.message ("Sorry",
                    "No stars detected.\nParameters may need adjusting.");
          }

          imframe.setCursor (GRIP.GRIP_DEFAULT_CURSOR);
        }
        catch (NumberFormatException ex)
        {
          String msg = "Config values missing for StarSegmenter parameters.\n" +
                       "Try deleting file grip.properties and re-running GRIP.";
          net.grelf.Util.warning (this, "Error", msg);
        }

        break;

      case "autosegnonsat":
        try
        {
          StarSegmenter.segment (imframe, Integer.parseInt (
                  Config.getValue ("StarSegmentDifference.value")),
                                          Integer.parseInt (
                  Config.getValue ("StarSegmentRadius.value")),
                  true); // Erode/dilate
          imframe.addHistory ("Segment non-saturated stars automatically");
          boolean includeSinglePixelBlobs = false;
          imframe.setCursor (GRIP.GRIP_WAIT_CURSOR);
          BlobMeasList allBlobMeases = imframe.detectBlobs (includeSinglePixelBlobs);

          if (null != allBlobMeases && allBlobMeases.size () > 0)
          {
        		blobMeases = new BlobMeasList (this.imframe);

            for (BlobMeas bm : allBlobMeases)
            {
              if (!bm.blob.isSaturatedInRegion (imframe))
              {
                blobMeases.add (bm);
              }
            }

            allBlobMeases.clear ();

            imframe.setBlobMeases (blobMeases);
            // Set reference in ImFrame so that measurements can be shown.

            reviseSettings ();
            imframe.getGP ().setDrawingMode (DrawingMode.BLOB_HOVER);
          }
          else
          {
            net.grelf.Util.message ("Sorry",
                    "No stars detected.\nParameters may need adjusting.");
          }

          imframe.setCursor (GRIP.GRIP_DEFAULT_CURSOR);
        }
        catch (NumberFormatException ex)
        {
          String msg = "Config values missing for StarSegmenter parameters.\n" +
                       "Try deleting file grip.properties and re-running GRIP.";
          net.grelf.Util.warning (this, "Error", msg);
        }

        break;

      case "blobs":
        imframe.addHistory ("Detect blobs");
        boolean includeSinglePixelBlobs = !net.grelf.Util.confirm (
                imframe, "Detect blobs", "Ignore 1-pixel blobs?");
        imframe.setCursor (GRIP.GRIP_WAIT_CURSOR);
        blobMeases = imframe.detectBlobs (includeSinglePixelBlobs);
        //		  imframe.setBlobMeases (blobMeases);
                // Set reference in ImFrame so that measurements can be shown.
        reviseSettings ();

        if (null == blobMeases || (0 == blobMeases.size ()))
        {
          net.grelf.Util.warning ("Sorry", "No blobs detected");
        }
        else
        {
          imframe.getGP ().setDrawingMode (DrawingMode.BLOB_HOVER);
        }

        imframe.setCursor (GRIP.GRIP_DEFAULT_CURSOR);
        break;

      case "blobbright":
        if (null == blobMeases)
        {
          net.grelf.Util.warning ("Sorry",
                  "Blob measurements not available. Detect again.");
        }
        else
        {
          imframe.addHistory ("Show brightest blobs");
          blobMeases.showBrightestBlobs (imframe);
        }

        break;

      case "blobdrawB":
        if (null == blobMeases)
        {
          net.grelf.Util.warning ("Sorry",
                  "Blob measurements not available. Detect again.");
        }
        else
        {
          imframe.addHistory ("Draw blob boundaries");
          blobMeases.drawBoundaries (imframe);
        }

        break;

      case "blobdrawR":
        if (null == blobMeases)
        {
          net.grelf.Util.warning ("Sorry",
                  "Blob measurements not available. Detect again.");
        }
        else
        {
          imframe.addHistory ("Draw blob regions");
          blobMeases.drawRegions (imframe);
        }

        break;

      case "blobhover":
        if (null == blobMeases)
        {
          net.grelf.Util.warning ("Sorry",
                  "Blob measurements not available. Detect again.");
        }
        else
        {
          imframe.getGP ().setDrawingMode (DrawingMode.BLOB_HOVER);
        }

        break;

      case "blobmeas":
        if (null == blobMeases)
        {
          net.grelf.Util.warning ("Sorry",
                  "Blob measurements not available. Detect again.");
        }
        else
        {
          if (blobMeases.size () == 0)
          {
            net.grelf.Util.warning ("Sorry", "No blobs detected");
          }
          else
          {
            imframe.addHistory ("View blob measurements");
            String response = (String) javax.swing.JOptionPane.showInputDialog
            (
              this.imframe, // Parent window
              "Measurement type?",					 // Prompt
              "Select measurements to show",			 // Title
              javax.swing.JOptionPane.QUESTION_MESSAGE,	 // Icon kind
              null,							 // No icon file
              new String []					 // Choices
                { "Area & brightness histograms",
                  "Width/height & circularity histograms",
                  "Colours (R/G v B/G)",
                  "Brightness v colour index",
                  "Brightness v mean radius",
                  "Peak level v mean radius",
                  "Most uncircular" },
              "Colours (R/G v B/G)"					// Default choice
            );

            if (null != response)
            {
              switch (response)
              {
                case "Area & brightness histograms":
                  new HistogramBlobArea (imframe, blobMeases).display ();
                  break;

                case "Brightness v colour index":
                  new ColourGraph (blobMeases, ColourGraph.GraphKind.BRIGHTNESS);
                  break;

                case "Colours (R/G v B/G)":
                  new ColourGraph (blobMeases, ColourGraph.GraphKind.COLOUR);
                  break;

                case "Brightness v mean radius":
                  new BlobGraph (blobMeases, BlobGraph.GraphKind.RADIUS);
                  break;

                case "Most uncircular":
                  blobMeases.showUncircular (imframe);
                  break;

                case "Peak level v mean radius":
                  new BlobGraph (blobMeases, BlobGraph.GraphKind.PEAK);
                  break;
                  
                case "Width/height & circularity histograms":
                  new HistogramBlobShape (imframe, blobMeases).display ();
                  break;

              }
            }
          }
        }

        break;

      case "blobsave":
        imframe.saveBlobMeasList ();
        break;

      case "clear":
        imframe.addHistory ("Clear overlay");
        imframe.getGP ().setDrawingMode (DrawingMode.UNDEFINED);
        break;

      case "drawoverlay":
        imframe.drawOverlay (true); // New image
        break;

      case "help":
        HelpMenu.help ("pages/seg_menu.html");
        break;

      case "split":
        splitStars ();
        break;

      case "threshold":
        imframe.addHistory ("Threshold");
        imframe.threshold ();
        break;

      case "warping":
        WarpingGrid grid = new WarpingGrid ();

        if (grid.make (this.blobMeases))
        {
          grid.save ();
        }

        break;
    }
	} // actionPerformed

  private void splitStars ()
  {
    net.grelf.image.Image srcIm = imframe.getImage ();
    net.grelf.image.Image starIm =
      net.grelf.image.ImageBase.createImage (
            srcIm.getWidth (), srcIm.getHeight (),
            srcIm.getNBands (), srcIm.getBitsPerChannel (),
            true, false); // fill with zeroes, not displayable
    net.grelf.image.Image backIm = srcIm.clone ();
    imframe.setCursor (imframe.IMFRAME_WAIT_CURSOR);

    try
    {
      int nDilations = 3;
      StarSegmenter.splitFromBackground (srcIm,
        Integer.parseInt (Config.getValue ("StarSegmentDifference.value")),
        Integer.parseInt (Config.getValue ("StarSegmentRadius.value")),
        nDilations, starIm, backIm);
      
      if (srcIm.getBitsPerChannel () > 32)
      {
        starIm.getRangeByRescanningDouble ();
        backIm.getRangeByRescanningDouble ();
      }
      else
      {
        starIm.getRangeByRescanning ();
        backIm.getRangeByRescanning ();
      }
      
      new ImFrame ("Stars", starIm);
      new ImFrame ("Background", backIm);
    }
    catch (net.grelf.image.IncompatibleImageException ex)
    {
      net.grelf.Util.logWarning ("Incompatible images for splitting stars");
    }

    imframe.setCursor (imframe.IMFRAME_DEFAULT_CURSOR);
  } // splitStars

	/** Calls reviseSettings (). Implementing MenuListener. */
	@Override
	public void menuSelected (javax.swing.event.MenuEvent ev)
	{
		reviseSettings ();
	} // menuSelected

	/** Implementing MenuListener but does nothing. */
	@Override
	public void menuDeselected (javax.swing.event.MenuEvent ev) {}

	/** Implementing MenuListener but does nothing. */
	@Override
	public void menuCanceled (javax.swing.event.MenuEvent ev) {}

} // SegmentationMenu